local aUF = LibStub("AceAddon-3.0"):GetAddon("ag_UnitFrames")
local plugin = aUF:NewModule("Rangecheck")
local newFrame, delFrame = aUF.newFrame, aUF.delFrame

plugin.defaultDisabledState = true

local pool = {}

local BS = {
	["Flash Heal"] = "Flash Heal",
	["Healing Touch"] = "Healing Touch",
	["Holy Light"] = "Holy Light",
	["Healing Wave"] = "Healing Wave",
	["Corruption"] = "Corruption",
	["Fear"] = "Fear",
	["Shadow Bolt"] = "Shadow Bolt",
	["Fireball"] = "Fireball",
	["Fire Blast"] = "Fire Blast",
}

local distanceCheckFunction
local distanceCheckFunctionLow
local running
do
	local _,class = UnitClass("player")
	if class == "PRIEST" then
		distanceCheckFunction = function(unit) return IsSpellInRange(BS["Flash Heal"], unit) == 1 or not UnitIsFriend("player", unit) end
	elseif class == "DRUID" then
		distanceCheckFunction = function(unit) return IsSpellInRange(BS["Healing Touch"], unit) == 1 or not UnitIsFriend("player", unit) end
	elseif class == "PALADIN" then
		distanceCheckFunction = function(unit) return IsSpellInRange(BS["Holy Light"], unit) == 1 or not UnitIsFriend("player", unit) end
	elseif class == "SHAMAN" then
		distanceCheckFunction = function(unit) return IsSpellInRange(BS["Healing Wave"], unit) == 1 or not UnitIsFriend("player", unit) end
	elseif class == "WARLOCK" then
		distanceCheckFunction = function(unit) return IsSpellInRange(BS["Corruption"], unit) == 1 or IsSpellInRange(BS["Shadow Bolt"], unit) == 1 or UnitIsFriend("player", unit) end
		distanceCheckFunctionLow = function(unit) return IsSpellInRange(BS["Fear"], unit) == 1 or UnitIsFriend("player", unit) end
	elseif class == "MAGE" then
		distanceCheckFunction = function(unit) return IsSpellInRange(BS["Fireball"], unit) == 1 or UnitIsFriend("player", unit) end
		distanceCheckFunctionLow = function(unit) return IsSpellInRange(BS["Fire Blast"], unit) == 1 or UnitIsFriend("player", unit) end
	else
		distanceCheckFunction = function(unit) return CheckInteractDistance(unit, 4) end
	end
end

function plugin:OnRegister()
	self:RegisterDefaults({
		units = {
			["**"] = {
				rangeCheck = false,
			},		
		},		
	})
end

function plugin:OnRegisterEvents(object, power)
	if plugin.db.profile.units[object.type].rangeCheck and not power then
		if not self.schedule then          
			self.schedule = aUF.StartTimer(self, "UpdatePool", 0.7)
		end
		pool[object] = true
	else
		pool[object] = nil
		if next(pool) == nil and self.schedule then
			self:CancelTimer(self.schedule)
			self.schedule = nil
		end		
	end
end

function plugin:OnObjectDisable(object)
	object.frame:SetAlpha(1)
end

function plugin:OnUpdateAll(object)
	if plugin.db.profile.units[object.type].rangeCheck and self:IsEnabled() then
		self.RangeCheck(object)
	else
		object.frame:SetAlpha(1)
	end
end

function plugin:OnEnable()
	for _,object in aUF:IterateUnitObjects() do
		if plugin.db.profile.units[object.type].rangeCheck then
			if self.unit then
				self:OnRegisterEvents(object)
			end
		end
	end
end

function plugin:OnDisable()
	if self.schedule then
		self:CancelTimer(self.schedule)
		self.schedule = nil
	end
	for _,object in aUF:IterateUnitObjects() do
		object.frame:SetAlpha(1)
	end
end

function plugin:UpdatePool()
	if next(pool) == nil and self.schedule then
		self:CancelTimer(self.schedule)
		self.schedule = nil
	end
	for object in pairs(pool) do
		self.RangeCheck(object)
	end
end

function plugin.RangeCheck(self)
	local opacity = 0.5
	if self.frame:IsVisible() then
		if distanceCheckFunction(self.unit) then
			if distanceCheckFunctionLow then
				if distanceCheckFunctionLow( self.unit ) then
					self.frame:SetAlpha(1)
				else
					self.frame:SetAlpha(opacity+(1-opacity)/2)
				end
			else
				self.frame:SetAlpha(1)
			end
		else
			self.frame:SetAlpha(opacity)
		end
	end
end
